// SPDX-License-Identifier: GPL-2.0+
/*
 * Copyright 2018-2020 RnD Center "ELVEES", JSC
 */

#include <linux/debugfs.h>
#include <linux/dma-mapping.h>
#include <linux/of.h>

#include "elcore50-pm.h"
#include "elcore50-reset.h"

#ifndef ELCORE50_NO_IRQS
static void elcore50_free_irqs(struct cluster_priv *drv_priv)
{
	int i, irq;

	for (i = 0; i < drv_priv->ncores; i++) {
		for (irq = 0; irq < drv_priv->nirqs; ++irq) {
			if (drv_priv->irqs[i * drv_priv->nirqs + irq] == 0)
				break;
			devm_free_irq(drv_priv->dev,
				      drv_priv->irqs[i * drv_priv->nirqs + irq],
				      &drv_priv->cores[i]);
		}
	}
	devm_kfree(drv_priv->dev, drv_priv->irqs);
}
#endif

int elcore50_probe(struct platform_device *pdev)
{
	struct cluster_priv *drv_priv;
	int ret, ncores;
#ifndef ELCORE50_MCOM03
	uint32_t tmp_read;
#endif
#ifndef ELCORE50_NO_IRQS
	int i;
#endif
	struct resource *res;

	drv_priv = devm_kzalloc(&pdev->dev, sizeof(struct cluster_priv),
				GFP_KERNEL);
	if (!drv_priv)
		return -ENOMEM;
	drv_priv->dev = &pdev->dev;

	ret = elcore50_reset_init(drv_priv);
	if (ret)
		return ret;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		ret = -ENOENT;
		dev_err_probe(&pdev->dev, ret, "Failed to get registers\n");
		goto reset_deinit;
	}

	ret = dma_set_mask_and_coherent(&pdev->dev, DMA_BIT_MASK(64));
	if (ret) {
		dev_err_probe(&pdev->dev, ret, "Failed to set DMAMASK\n");
		goto reset_deinit;
	}

	drv_priv->regs = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(drv_priv->regs)) {
		ret = PTR_ERR(drv_priv->regs);
		dev_err_probe(&pdev->dev, ret, "Failed to map registers\n");
		goto reset_deinit;
	}
	res = platform_get_resource(pdev, IORESOURCE_MEM, 1);
	if (!res) {
		ret = -ENOENT;
		dev_err_probe(&pdev->dev, ret, "Failed to get memory\n");
		goto reset_deinit;
	}

	drv_priv->mem = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(drv_priv->mem)) {
		ret = PTR_ERR(drv_priv->mem);
		dev_err_probe(&pdev->dev, ret, "Failed to map registers\n");
		goto reset_deinit;
	}

	platform_set_drvdata(pdev, drv_priv);

	ret = elcore50_clock_init(drv_priv);
	if (ret)
		goto err_clock;

#ifdef ELCORE50_MCOM03
	ncores = 1;
#else
	tmp_read = ioread32(drv_priv->regs + DSP_IDR);
	ncores = (tmp_read & IDR_COCNT_MASK) >> IDR_COCNT_OFFT;
	WARN_ON((tmp_read & IDR_VER_MASK) >> IDR_VER_OFFT != 0x32);
	WARN_ON((tmp_read & IDR_REV_MASK) >> IDR_REV_OFFT != 0x66);
#endif
	ret = elcore50_cores_init(drv_priv, ncores);
	if (ret)
		goto err_clock;
#ifndef ELCORE50_NO_IRQS
	drv_priv->nirqs = of_property_count_elems_of_size(
		drv_priv->dev->of_node, "interrupts", sizeof(int));
	drv_priv->irqs = devm_kcalloc(&pdev->dev, drv_priv->nirqs,
				      sizeof(unsigned int),
				      GFP_KERNEL | __GFP_ZERO);
	if (!drv_priv->irqs) {
		ret = -ENOMEM;
		goto err_device;
	}
	drv_priv->nirqs /= ncores;
	for (i = 0; i < drv_priv->nirqs * ncores; i++) {
		drv_priv->irqs[i] = platform_get_irq(pdev, i);
		if (drv_priv->irqs[i] < 0) {
			ret = drv_priv->irqs[i];
			dev_err_probe(&pdev->dev, ret,
				      "Failed to get interrupt\n");
			goto err_irq;
		}
		ret = devm_request_irq(&pdev->dev, drv_priv->irqs[i],
				       elcore50_irq, IRQF_SHARED, "elcore50",
				       &drv_priv->cores[i / drv_priv->nirqs]);

		if (ret) {
			ret = -EINVAL;
			dev_err_probe(&pdev->dev, ret,
				      "Failed to get interrupt resource\n");
			goto err_irq;
		}
	}
#endif
	pm_runtime_suspend(drv_priv->dev);

	dev_info(drv_priv->dev, "%d ELcore-50 cores initialized\n", ncores);
	return 0;
#ifndef ELCORE50_NO_IRQS
err_irq:
	while (i) {
		devm_free_irq(drv_priv->dev, drv_priv->irqs[i - 1],
			      &drv_priv->cores[(i - 1) / drv_priv->nirqs]);
	}
	devm_kfree(drv_priv->dev, drv_priv->irqs);
#endif
err_device:
	elcore50_cores_destroy(drv_priv);
err_clock:
	elcore50_clock_destroy(drv_priv);
reset_deinit:
	elcore50_reset_fini(drv_priv);
	return dev_err_probe(&pdev->dev, ret, "elcore50 init failed\n");
}

int elcore50_remove(struct platform_device *pdev)
{
	struct cluster_priv *drv_priv = platform_get_drvdata(pdev);

#ifndef ELCORE50_NO_IRQS
	elcore50_free_irqs(drv_priv);
#endif
	elcore50_cores_destroy(drv_priv);
	elcore50_clock_destroy(drv_priv);
	elcore50_reset_fini(drv_priv);
	return 0;
}
